# Aproximarea Functiilor Neliniare folosind Retele Neuronale
# Ciprian Pungila
# Univ. Politehnica, Fac. de Automatica si Calculatoare
# Anul III
# Version 2.0
# Compiled with Python 2.4, wxPython & NumPy

from wxPython.wx               import *
from wxPython.lib.wxPlotCanvas import *
from wxPython.lib              import wxPlotCanvas
from wxPython.lib.rcsizer      import RowColSizer

import random
import math
import Numeric

def sign(x):
    if x >= 0 :
        return 1
    else:
        return -1

def exp(x):
    if x > 20 :
        return math.exp(20)
    else:
        if x < -20:
            return math.exp(-20)
        else:
            return math.exp(x)

def _InitObjects():
    # this initial representation is only for intro purposes,
    # it does not affect the program in other ways
    
    # 100 points sin function, plotted as green circles
    data1 = 2.*Numeric.pi*Numeric.arange(200)/200.
    data1.shape = (100, 2)
    data1[:,1] = Numeric.sin(data1[:,0])
    markers1 = PolyMarker(data1, color='green', marker='circle',size=1)

    # 50 points cos function, plotted as red line
    data1 = 2.*Numeric.pi*Numeric.arange(100)/100.
    data1.shape = (50,2)
    data1[:,1] = Numeric.cos(data1[:,0])
    lines = PolyLine(data1, color='red')

    # A few more points...
    pi = Numeric.pi
    markers2 = PolyMarker([(0., 0.), (pi/4., 1.), (pi/2, 0.),
                          (3.*pi/4., -1)], color='blue',
                          fillcolor='green', marker='cross')

    return PlotGraphics([markers1, lines, markers2])

    # the class uses to build our neural network
class network:

    # this is the default constructor
    def __init__(self, mainwin, a, b, nr_centers, nr_max_centers, nr_max_epochs, learning_rate, parameters_values, nr_examples_from_set, learning_tolerance,
                 adjusting_coefficient, magnitude, function):

        self.mainwin = mainwin
        self.a = a
        self.b = b
        self.linit = nr_centers
        self.l = nr_max_centers
        self.epochs = nr_max_epochs
        self.n = learning_rate
        self.deltamic = parameters_values
        self.ex = nr_examples_from_set
        self.e = learning_tolerance
        self.alfa = adjusting_coefficient
        self.magnitude = magnitude
        self.function = self.mainwin.functions[function] # the function to estimate

        self.error = 1000 # currently calculated error
        self.h = float(self.b-self.a) / self.ex # this is the actual step

        random.seed()

        self.x = [self.a + (i-1)*self.h for i in range(1, self.ex+1, 1)]
        self.d = [self.function(self.x[i-1]) + self.magnitude*(2*random.random() - 1) for i in range(1, self.ex+1, 1)]
        self.c = [self.x[random.randint(0,self.ex-1)] for i in range(1, 1000, 1)]
        self.w0 = 0
        self.k = self.linit
        self.w = [random.random() for i in range(1, 1000, 1)]
        self.delta = [random.random() for i in range(1, 1000, 1)]
        self.y = [random.random() for i in range(1, 1000, 1)]
        

    # each time we call an instance of the class (e.g. network()), this method is called
    def __call__(self):
        return

    # compute estimated function value in xx using neural network
    def compute_estimated(self, x):
        rez = -self.w0
        for j in range(1, self.k+1, 1):
            if self.delta[j-1] != 0:
                rez = rez + self.w[j-1]*exp(-(x-self.c[j-1])*(x-self.c[j-1])/(2*self.delta[j-1]*self.delta[j-1]))
        return rez

    # learning function
    def learn(self):

        # determinarea centrelor
        t = 0
        kstea = 0
        while (t <= self.l) and (self.n >= self.e):
            for i in range(1, self.ex+1, 1):
                kstea = 1
                for j in range(1, self.k+1, 1):
                    if abs(self.x[i-1]-self.c[kstea-1]) > abs(self.x[i-1]-self.c[j-1]):
                        kstea = j
                if abs(self.x[i-1]-self.c[kstea-1] < self.deltamic):
                    self.c[kstea-1] = self.c[kstea-1] + self.n*(self.x[i-1]-self.c[kstea-1])
                else:
                    self.k = self.k + 1
                    self.c[self.k] = self.x[i-1]
            t = t + 1
            self.n = self.n * exp(-self.alfa * math.log(t))

        # determinarea largimilor campurilor receptive
        for t in range(1, self.k+1, 1):
            if (t != 1):
                i = 1
            else:
                i = 2
            for it in range(1, self.k+1, 1):
                if (abs(self.c[t-1] - self.c[it-1]) < abs(self.c[t-1]-self.c[i-1])) and (t != it):
                    i = it
            j = 1
            while (j == t) or (j == i):
                j = j+1
            for it in range(1, self.k+1, 1):
                if (abs(self.c[t-1]-self.c[it-1])<abs(self.c[t-1]-self.c[j-1])) and (it!=i) and (t!=it):
                    j = it
            self.delta[t-1] = (abs(self.c[t-1]-self.c[i-1])+abs(self.c[t-1]-self.c[j-1]))/2

        # determinarea parametrilor W
        t = 0
        self.p = 0
        self.w0 = random.random() - 0.5
        for i in range(1, self.k+1, 1):
            self.w[i-1] = random.random() - 0.5

        # show dialog
        dlg = wxProgressDialog("Antrenare retea",
                       "Reteaua este antrenata",
                       self.epochs,
                       self.mainwin, # instance of type (wxWindow *)
                       wxPD_CAN_ABORT | wxPD_APP_MODAL | wxPD_AUTO_HIDE | wxPD_ELAPSED_TIME | wxPD_REMAINING_TIME | wxPD_ESTIMATED_TIME)
        keepGoing = true
        while keepGoing and (t <= self.epochs) and (abs(self.error) > self.e):
            # start calculus
            for i in range(1, self.ex+1, 1):
                self.y[i-1] = -self.w0
                for j in range(1, self.k+1, 1):
                    self.y[i-1] = self.y[i-1]+self.w[j-1]*exp(-(self.x[i-1]-self.c[j-1])*(self.x[i-1]-self.c[j-1])/(2*self.delta[j-1]*self.delta[j-1]))
                for j in range(1, self.k+1, 1):
                    self.w[j-1] = self.w[j-1]+self.n*(self.d[i-1]-self.y[i-1])*exp(-(self.x[i-1]-self.c[j-1])*(self.x[i-1]-self.c[j-1])/(2*self.delta[j-1]*self.delta[j-1]))
                    self.w0 = self.w0 - self.n*(self.d[i-1]-self.y[i-1])
            t = t + 1
            # compute error here
            self.error = 0
            for i in range(1, self.ex+1, 1):
                self.error = self.error + (self.d[i-1]-self.y[i-1])*(self.d[i-1]-self.y[i-1])
            self.error = self.error / self.ex

            # update probes
            self.p += 1

            self.mainwin.error_pg.add_point(self.p, self.error) # add point for redrawing the error graph
            self.mainwin.pg.make_funcs_aux() # add point for redrawing the estimated function graph
 
            self.mainwin.plot.redraw()
            self.mainwin.error_plot.redraw()
            self.mainwin.update_probes()
            keepGoing = dlg.Update(min(self.p, self.epochs))
        # now everything has been calculated, time to show the
        # output graphs
        dlg.Destroy()
        

class nn_errorplotcanvas(PlotCanvas):
    def __init__(self, parent):
        PlotCanvas.__init__(self, parent)

class nn_errorplotgraphics(PlotGraphics):
    point_q = 50.0
    def __init__(self):
        self.objects = []
        self.points = []
        self.last_scale = [1,1]
        self.last_shift = [0,0]

        self.objects.append( PolyLine(self.points, color="red") )

    def add_point(self, x, y):
        self.points.append( [x,y] )
        self.objects = []
        self.objects.append( PolyLine(self.points, color="red") )

    def scaleAndShift(self, scale=1, shift=0):
        self.last_scale = scale
        self.last_shift = shift
        PlotGraphics.scaleAndShift(self, scale, shift)
        
class nn_plotcanvas(PlotCanvas):
    def __init__(self, parent, error_plot):
        PlotCanvas.__init__(self, parent)
        self.error_plot = error_plot
        self.parent = parent
        
        
class nn_plotgraphics(PlotGraphics):
    point_q = 50.0
    # default constructor
    def __init__(self, network, target, error_pg = None):
        self.objects = [] 
        self.run = 0
        self.target = target
        self.network = network
        self.error_pg = error_pg
        self.make_funcs()

    # make the functions to plot - auxiliary function  
    def make_funcs_aux(self):
        delme = []
        for o in self.objects:
            if isinstance(o, PolyLine):
                delme.append( o )
        for o in delme:
            self.objects.remove( o )
            
        net_points = []
        target_points = []
        step_size = abs(self.network.a-self.network.b)/self.point_q
        start = self.network.a
        for steps in range(self.point_q):
            x = start + steps*step_size
            y = self.network.compute_estimated(x)
            net_points.append( (x, y) )
            y = self.target(x)
            target_points.append( (x,y) )
        self.objects.append(PolyLine(net_points, color="blue"))
        self.objects.append(PolyLine(target_points, color="green"))

    # make the functions to plot    
    def make_funcs(self):
        delme = []
        for o in self.objects:
            if isinstance(o, PolyLine):
                delme.append( o )
        for o in delme:
            self.objects.remove( o )
            
        net_points = []
        target_points = []
        step_size = abs(self.network.a-self.network.b)/self.point_q
        start = self.network.a
        for steps in range(self.point_q):
            x = start + steps*step_size
            y = self.network.compute_estimated(x)
            net_points.append( (x, y) )
            y = self.target(x)
            target_points.append( (x,y) )
        self.objects.append(PolyLine(net_points, color="blue"))
        self.objects.append(PolyLine(target_points, color="green"))

    # scale and shift the graphic
    def scaleAndShift(self, scale=1, shift=0):
        self.last_scale = scale
        self.last_shift = shift
        PlotGraphics.scaleAndShift(self, scale, shift)

        
class main_window(wxFrame):
    def __init__(self, parent, id):
        wxFrame.__init__(self, parent, -1, "Aproximarea functiilor neliniare cu retele neuronale - Versiunea 2.0", size = (800, 500),
                         style=wxDEFAULT_FRAME_STYLE|wxNO_FULL_REPAINT_ON_RESIZE)


        # ------------------------------------------------------------------------------------
        # Set up menu bar for the program.
        # ------------------------------------------------------------------------------------
        self.mainmenu = wxMenuBar()   # Create menu bar.
        mainwindow = self

        menu=wxMenu()                                 # Make a menu (will be the Project menu)

        exitID=wxNewId()
        menu.Append(exitID, '&Nou...', 'Creeaza o retea noua')
        EVT_MENU(self, exitID, self.OnNew)
        
        exitID=wxNewId()
        menu.Append(exitID, '&Iesire', 'Iesire din program')
        EVT_MENU(self, exitID, self.OnExit)

        self.mainmenu.Append (menu, '&Fisier')       # Add the project menu to the menu bar.

        menu=wxMenu()                                 # Make a menu (will be the Project menu)

        exitID=wxNewId()
        menu.Append(exitID, '&Despre', 'Informatii')
        EVT_MENU(self, exitID, self.OnAbout)
        
        self.mainmenu.Append (menu, '&Ajutor')       # Add the project menu to the menu bar.



        self.SetMenuBar (self.mainmenu)               # Attach the menu bar to the window.
        self.CreateStatusBar()


  

        # ------------------------------------------------------------------------------------
        # Create horizontal the splitter window.
        # ------------------------------------------------------------------------------------
        hsid = wxNewId()
        self.hsplitter = hsplitter = wxSplitterWindow (self, hsid, style=wxNO_3D|wxSP_3D)
        #hsplitter.SetMinimumPaneSize (1)
        # ------------------------------------------------------------------------------------
        # Create the configuration panel
        # ------------------------------------------------------------------------------------

        self.cpanel = cpanel = wxPanel(hsplitter, -1)

        # ------------------------------------------------------------------------------------
        # Create the vertical splitter window.
        # ------------------------------------------------------------------------------------
        splitter = wxSplitterWindow (hsplitter, -1, style=wxNO_3D|wxSP_3D)
        splitter.update_probes = self.update_probes
        splitter.SetMinimumPaneSize (1)
        

        hsplitter.SplitHorizontally (self.cpanel, splitter)
        # ------------------------------------------------------------------------------------
        # Create the editor on the right.
        # ------------------------------------------------------------------------------------
        self.error_plot = win = nn_errorplotcanvas(splitter)
        win.draw(_InitObjects(),'automatic','automatic')
        
        # ------------------------------------------------------------------------------------
        # Create the plot canvas
        # ------------------------------------------------------------------------------------
        tID = wxNewId()
        self.plot =  win = nn_plotcanvas(splitter, self.error_plot)
        win.draw(_InitObjects(),'automatic','automatic')
        

        # ------------------------------------------------------------------------------------
        # Install the tree and the editor.
        # ------------------------------------------------------------------------------------
        splitter.SplitVertically (self.plot, self.error_plot)
        splitter.SetSashPosition (420, true)

        # place controls
        gridsz = RowColSizer()
        

        # new network
        line = 1

        # A and B for the definition domain
        c = wxStaticText(cpanel, -1, " Start: ")
        gridsz.Add( c, pos=(line,1))
        self.lrange_vals = ["0","-0.5","-1","-1.5","-2","-3","-5","-10","-50","-100"]
        self.lrange = "-5"
        cbID = wxNewId()
        c = wxComboBox(cpanel, cbID, self.lrange, choices=self.lrange_vals,
                                     size=(100,-1), style=wxCB_DROPDOWN)
        gridsz.Add( c, pos=(line,2))                                     
        EVT_COMBOBOX(self, cbID, self.OnComboLrange)       

        c = wxStaticText(cpanel, -1, " Stop: ")
        gridsz.Add( c, pos=(line,3))

        self.rrange_vals = ["0","0.5","1","1.5","2","3","5","10","50","100"]
        self.rrange = "5"
        cbID = wxNewId()
        c = wxComboBox(cpanel, cbID, self.rrange, choices=self.rrange_vals,
                                     size=(100,-1), style=wxCB_DROPDOWN)
        gridsz.Add( c, pos=(line,4))                                     
        EVT_COMBOBOX(self, cbID, self.OnComboRrange)       

        # initial number of centers
        line += 1
        c = wxStaticText(cpanel, -1, " Numarul initial de centre: ")
        gridsz.Add( c, pos=(line,1))                   

        self.nr_centers = c = wxTextCtrl(cpanel, -1, "2")
        gridsz.Add( c, pos=(line,2))

        # maximum number of centers
        line += 1
        c = wxStaticText(cpanel, -1, " Numarul maxim de centre: ")
        gridsz.Add( c, pos=(line,1))                   

        self.nr_max_centers = c = wxTextCtrl(cpanel, -1, "50")
        gridsz.Add( c, pos=(line,2))

        # maximum number of epochs
        line += 1
        c = wxStaticText(cpanel, -1, " Numarul maxim de epoci: ")
        gridsz.Add( c, pos=(line,1))                   

        self.nr_max_epochs = c = wxTextCtrl(cpanel, -1, "5000")
        gridsz.Add( c, pos=(line,2))

        # learning rate
        line += 1
        c = wxStaticText(cpanel, -1, " Rata de invatare: ")
        gridsz.Add( c, pos=(line,1))                   

        self.learning_rate = c = wxTextCtrl(cpanel, -1, "0.1")
        gridsz.Add( c, pos=(line,2))

        # parameters' values
        line += 1
        c = wxStaticText(cpanel, -1, " Valoarea parametrilor: ")
        gridsz.Add( c, pos=(line,1))                   

        self.parameters_values = c = wxTextCtrl(cpanel, -1, "0.1")
        gridsz.Add( c, pos=(line,2))

        # number of examples from set
        line += 1
        c = wxStaticText(cpanel, -1, " Numar exemple din set: ")
        gridsz.Add( c, pos=(line,1))                   

        self.nr_examples_from_set = c = wxTextCtrl(cpanel, -1, "20")
        gridsz.Add( c, pos=(line,2))

        # new network !
        line = 1
        cbID = wxNewId()
        c = wxButton(cpanel , cbID, "Retea Noua")
        gridsz.Add( c, pos=(line,5))                                     
        EVT_BUTTON(self, cbID, self.OnNew)

        # training controls
        cbID = wxNewId()
        c = wxButton(cpanel, cbID, "Antrenare")
        gridsz.Add( c, pos=(line,6))                                                                
        EVT_BUTTON(self, cbID, self.OnTrain)

        # adjusting coefficient
        line += 1
        c = wxStaticText(cpanel, -1, " Coef. de ajust. a centrilor: ")
        gridsz.Add( c, pos=(line,3))                   

        self.adjusting_coefficient = c = wxTextCtrl(cpanel, -1, "1")
        gridsz.Add( c, pos=(line,4))

        # magnitude
        line += 1
        c = wxStaticText(cpanel, -1, " Magnitudine zgomot: ")
        gridsz.Add( c, pos=(line,3))                   

        self.magnitude = c = wxTextCtrl(cpanel, -1, "0.1")
        gridsz.Add( c, pos=(line,4))

        # the main function
        line += 1
        c = wxStaticText(cpanel, -1, " Functia: ")
        gridsz.Add( c, pos=(line,3))
        self.functions = { "1 (constanta)": lambda x: 1,
                            "x": lambda x: x,
                            "sign": sign,
                            "sinus": lambda x: math.sin(x),
                            "cosinus": lambda x: math.cos(x),
                            "x^2": lambda x: x*x,
                            "x^3": lambda x: x*x*x,
                            "x*sin x": lambda x: x*math.sin(x),
                            "x*sin(5*x)": lambda x: x*math.sin(5*x),
                            "x*x+1": lambda x: x*x+1,
                            "10*x*x+1": lambda x: 10*x*x+1
                            }
        self.function = "sinus"
        
        cbID = wxNewId()
        c = wxComboBox(cpanel, cbID, self.function, choices=self.functions.keys(),
                                     size=(100,-1), style=wxCB_DROPDOWN)
        gridsz.Add( c, pos=(line,4))                                     
        EVT_COMBOBOX(self, cbID, self.OnComboFunction)                                     

        # learning tolerance
        line += 1
        c = wxStaticText(cpanel, -1, " Toleranta la invatare: ")
        gridsz.Add( c, pos=(line,3))                   

        self.learning_tolerance = c = wxTextCtrl(cpanel, -1, "0.001")
        gridsz.Add( c, pos=(line,4))

        # add function response to manual sample
        line += 1
        c = wxStaticText(cpanel, -1, " Dati o valoare: ")
        gridsz.Add( c, pos=(line,3))                   

        self.point_to_compute = c = wxTextCtrl(cpanel, -1, "0")
        gridsz.Add( c, pos=(line,4))

        cbID = wxNewId()
        c = wxButton(cpanel, cbID, "Calculeaza")
        gridsz.Add( c, pos=(line,5))                                                                
        EVT_BUTTON(self, cbID, self.OnCalculate)

        c = wxStaticText(cpanel, -1, " Valoarea reala: ")
        gridsz.Add( c, pos=(line,6))                

        self.real_value = c = wxTextCtrl(cpanel, -1, "0", style = wxTE_READONLY)
        gridsz.Add( c, pos=(line,7))

        line += 1
        c = wxStaticText(cpanel, -1, " Valoarea calculata: ")
        gridsz.Add( c, pos=(line,6))                

        self.calculated_value = c = wxTextCtrl(cpanel, -1, "0", style = wxTE_READONLY)
        gridsz.Add( c, pos=(line,7))

        # probes
        line = 3
        c = wxStaticText(cpanel, -1, " Pasul curent: ")
        gridsz.Add( c, pos=(line,6))                   

        self.current_step = c = wxTextCtrl(cpanel, -1, "0", style = wxTE_READONLY)
        gridsz.Add( c, pos=(line,7))                  

        line += 1
        c = wxStaticText(cpanel, -1, " Eroarea curenta: ")
        gridsz.Add( c, pos=(line,6))                   

        self.current_error = c = wxTextCtrl(cpanel, -1, "0", style = wxTE_READONLY)
        gridsz.Add( c, pos=(line,7))                  


        self.cpanel.SetSizer(gridsz)
        self.cpanel.SetAutoLayout(true)
        self.Show(true)
        hsplitter.SetSashPosition(185)

    def OnComboHidden(self, evt):
        self.layers = int(evt.GetString())

    def OnComboFunction(self, evt):
        self.function = evt.GetString()

    def OnComboAlpha(self, evt):
        self.eta = evt.GetString()

    def OnComboRrange(self, evt):
        self.rrange = evt.GetString()

    def OnComboLrange(self, evt):
        self.lrange = evt.GetString()

    def OnComboTrain(self, evt):
        return
        
    def OnExit(self, evt):
        self.Close(true)

    def OnAbout(self, event):
        dlg = wxMessageDialog(self, "Aproximarea functiilor neliniare cu retele neuronale\nVersiunea 2.0\n\nCiprian Pungila, Fac. de Automatica si Calculatoare\n\nE-mail: cipy@email.com",
                              "Despre", wxOK | wxICON_INFORMATION)
        dlg.ShowModal()
        dlg.Destroy()


    # event generated when the neural network is being trained    
    def OnTrain(self, evt):
            self.network.learn()
            self.plot.redraw()
            self.error_plot.redraw()
            self.update_probes()

        
    # event generated when we need to compute an individual value    
    def OnCalculate(self, evt):
        try:
            to_compute = float(self.point_to_compute.GetValue())
            self.real_value.SetValue( str(self.pg.target(to_compute)) )
            self.calculated_value.SetValue( str(self.pg.network.compute_estimated(to_compute)) )
        except:
            msgdlg = wxMessageDialog(self, 'A aparut o eroare la calcul! Probabil un overflow matematic?',
                          'Eroare la calcul', wxOK | wxICON_EXCLAMATION)

            msgdlg.ShowModal()
            msgdlg.Destroy()




    # event generated when creating a new neural network
    def OnNew(self, evt):
        self.network = network(self, # mainwin
                               int(self.lrange), #a
                               int(self.rrange), #b
                               int(self.nr_centers.GetValue()), # number of centers
                               int(self.nr_max_centers.GetValue()), # max. number of centers
                               int(self.nr_max_epochs.GetValue()), # max. number of epochs
                               float(self.learning_rate.GetValue()), # learning rate
                               float(self.parameters_values.GetValue()), # parameters' values
                               int(self.nr_examples_from_set.GetValue()), # number of examples from set
                               float(self.learning_tolerance.GetValue()), # learning tolerance
                               int(self.adjusting_coefficient.GetValue()), # adjusting coefficient
                               float(self.magnitude.GetValue()), # magnitude
                               function = self.function) # function to simulate

        # create the necessary classes
        self.error_pg = nn_errorplotgraphics()
        self.pg = nn_plotgraphics(self.network, self.functions[self.function], error_pg = self.error_pg)        

        # draw the graphs
        self.plot.draw( self.pg, 'automatic', 'automatic' )
        y = 1.0
        self.error_pg.add_point(0, y)
        self.error_plot.draw( self.error_pg, 'automatic', 'automatic' )        


    def update_probes(self):
        self.current_error.SetValue( str(self.network.error) ) # updating the error editfield
        self.current_step.SetValue( str(self.network.p) )     # and the number of steps

class App(wxApp):
    def OnInit(self):
        frame = main_window(None, -1)
        self.SetTopWindow(frame)
        return true

app = App(0)
app.MainLoop()

